/* contour.prg
** 
** Generate the indifference curve/UOF plot; saves in countour.eps
** Copyright (c) 2006 Branislav L. Slantchev (slantchev@ucsd.edu)
*/
new;
library pgraph,co;
graphset;

// solution lies on MC
delta = 0.90; ps_fname = "../figures/contour.ps"; 
// makes solution lie on CR
//delta = 0.98; ps_fname = "../figures/contour2.ps"; 
// makes solution lie on CR/MC intersect
//delta = 0.95; ps_fname = "../figures/contour3.ps"; 
T = 4;
w_h = 0.80;
W = 1.25;
e = 0.25;
b = 0.85;
c = 1.08;

//b = 0.97;
//c = 1.015;


proc(0)=which_gt(x,y,sx,sy);
    local s;
    if( x > y ); s = " is greater than ";
    elseif( x < y ); s = " is smaller than ";
    else; s = " is equal to ";
    endif;
    print /flush sx "=" x s y "=" sy;
endp;

proc calc_ah();
    retp( (b.*(W+w_h)./c).^(1./(c-b)) );
endp;

proc calc_a3();
    local dT;
    dt = delta^T;
    retp( ((1-dT)*(W+w_h)/(1-dT+(1-delta)*(1+e*(1-dT))))^(1/(c-b)) );
endp;

proc calc_CR(a);
    local dT;

    dT = delta^T;
    retp( ((1-dT).*(W+w_h-a.^(c-b)))./(1+(1-dT).*e) );
endp;

proc calc_MC(a);
    local dTp1;

    dTp1 = delta^(T+1);
    retp( ((1-dTp1).*a.^(c-b)-(delta-dTp1).*(W+w_h))./(1-delta.*(1+e.*(1-delta.^T))) );
endp;

proc calc_UOF(a);
    retp( (W+w_h-c.*a.^(c-b)./b)./e );
endp;

// 'a' can be Nx1 vector and 'minw' can be 1xK
// returns NxK matrix of payoffs
proc hpay(a,minw);
    local rv;
    rv = zeros(rows(a),cols(minw));
    if( rows(minw) > 1 or cols(a) > 1 );
        print /flush "hpay: matrices not conformable.";
        end;
    endif;
    for r (1,rows(a),1);
        rv[r,0] = (W-e*minw+w_h).*a[r].^b-a[r].^c;
    endfor;
    retp( rv );
endp;

proc gdp(x);
    local a,minw;
    a = x[1];
    minw = x[2];
    retp( -b*(W-e*minw+w_h)*a^(b-1)+c*a^(c-1) | e*a^b );
endp;

proc ineqp(x);
    local a, minw;
    a = x[1];
    minw = x[2];
    retp(  minw-calc_CR(a) | minw-calc_MC(a) );
    //retp( minw-calc_MC(a) ); // try MC only
    //retp( minw-calc_CR(a) ); // try CR only
endp;

proc neg_hpay(x);
    retp( -hpay(x[1],x[2]) );
endp;


/*
*****************************************************************
** MAIN
*****************************************************************
*/
format /rd 6,4;

max_e = (1-delta)/(delta*(1-delta^T));
print /flush "max e = " max_e;;
if( e > max_e ); print /flush "e = " e "is invalid"; end;
else; print /flush "";
endif;

const_TR = 1/(1-delta*(1+e*(1-delta^T)));
const_P = (1+e*(1-delta^T))/(1-delta^T);
const_R = (1-delta^(T+1))*const_TR;
const_S = delta*(1-delta^T)*const_TR;
K2 = (1+const_P*const_R)/(1+const_P*const_S);
K1 = (1+const_P*const_R)/((1+const_P*const_S)*(1+e));

which_gt(c/b,K1,"c/b","K1");
which_gt(c/b,K2,"c/b","K2");

if( c/b < K1 );
    print /flush "MC is binding";
    the_a = (b*(W+w_h)/(c*(1+e)))^(1/(c-b));
    the_w = calc_MC(the_a);
    print /flush "a* = " the_a "\b; w* = " the_w "\b; CR(a*) = " calc_CR(the_a);
elseif( c/b >= K1 and c/b <= K2 );
    print /flush "Solution is at the intersection";
    the_a = ((W+w_h)*(1+const_P*const_S)/(1+const_P*const_R))^(1/(c-b));
    the_w = calc_MC(the_a);
    print /flush "a* = " the_a "\b; w* = " the_w "\b; CR(a*) = " calc_CR(the_a);
else;
    print /flush "CR is binding";
    the_a = (b*(W+w_h)/c)^(1/(c-b));
    the_w = calc_CR(the_a);
    print /flush "a* = " the_a "\b; w* = " the_w "\b; MC(a*) = " calc_MC(the_a);
endif;

//end;

coset;
_co_IneqProc = &ineqp;
_co_GradProc = &gdp;
_co_DirTol = 1e-8;
_co_GradCheckTol = 1e-5;
_co_DirTol = 1e-7;
_co_Bounds = (1~1e256 | 0~w_h);
__output = 0;
// make sure start value admits a solution!
// sometimes optimizer gets confused
//start = (7~0.7);
start = (the_a~the_w);
trap 4;
{x,pay,g,ret} = co(&neg_hpay,start);
if( ret /0 );
    call coprt(x,pay,g,ret);
    //end;
endif;
print /flush 
  "(a*,w*) = (" x[1] "\b," x[2] "\b)"
  "\npayoff  = " -pay;
  "MC(a*) = " calc_MC(x[1]) "\b; CR(a*) = " calc_CR(x[1]);

// store coords for (a,w) at SPE
spe_x = x[1];
spe_y = x[2];

minw = seqas(0,w_h,51);
a = seqas(1,round(the_a*2),201);
mc = calc_MC(a);
cr = calc_CR(a);
uof = calc_UOF(a);
payoff = hpay(a,minw');

_pdate="";
_plwidth = 25;
_pltype = { 1 3 6 };
_pmcolor = { 0, 0, 0, 0, 0, 0, 0, 0, 15 };
_pcolor = 4;
_pqg_ps_WhiteBecomes = 15; // stop the postscript routines from converting white
_ptek = "contour.tkf";
xlabel("Contribution Level (a)");
ylabel("Smallest Contributor (w)");
xtics(1,maxc(a),1,0);
ytics(0,w_h,.10,0);
ztics(0,2.4,.15,0);

_pmsgstr = "MC]1[(a)\000CR]1[(a)\000UOF(a)\000SPE";
if( maxc(mc) < w_h ); mc_x = 11.5; mc_y = maxc(mc);
else;
    mc_x = a[minindc(abs(mc-w_h))+1];
    mc_y = w_h - .05;
endif;
uof_x = a[minindc(abs(uof-w_h))]-1.25;

// x,y,type,height,color,1,thickness
_psym = (spe_x~spe_y~8~6~0~1~0);
// x1,y1,x2,y2,head length:width,size of head,01,color,1,6,thickness
spe_x1 = spe_x-1.5;
spe_y1 = spe_y+.15;
_parrow = (spe_x1~spe_y1~(spe_x-.2)~(spe_y+.02)~3~.2~01~0~1~6~15);

// x,y,char height,0,1,color,font thickness
_pmsgctl = 
(    mc_x~mc_y~.15~0~1~0~10
  |  1.3~cr[1]~.15~0~1~0~10 
  | uof_x~mc_y~.15~0~1~0~10 
  | spe_x1-.3~spe_y1+.01~.15~0~1~0~10
);


begwind;
window(2,2,0); // one graphic panel
xsize = 9;
ysize = 6.855;
makewind(xsize,ysize,0,0,0);
makewind(xsize,ysize,0,0,1);
setwind(5);
contour(a',minw,payoff');
setwind(6);
xy(a,mc~cr~uof);
endwind;
call tkf2ps(_ptek,ps_fname);
