/*
** Copyright (c) 2006 Branislav L. Slantchev (slantchev@ucsd.edu)
*/
new;
library co,pgraph;
coset;
eqSolveset;
graphset;

#include params.dec;

// taking (c-b) small makes MC slope very small
b = 0.80;
c = 1.05;
T = 2;
// increasing delta shifts MC/CR down
delta = 0.99;

__DebugOutput = 1;  // if /= 0, print optimization messages
__DebugExtraStuff = 0; // if /=0, print intermediate calls to optimizer

w_h = 0.86;

// takes KxN matrix of W_dists and returns KxN matrix
// of coalition sizes such that if wk = W_dists[r,c], then
// C = retval[r,c], where C is sum of states with size >= wk
// given state system r 
proc coalitions(W_dists);
    local C_sizes,mask,wk;

    C_sizes = zeros(rows(W_dists),cols(W_dists));
    for ir (1, rows(W_dists), 1);    
        for ic (1, cols(W_dists), 1);
            // set wk and pick out wi >= wk only
            wk = W_dists[ir,ic];
            mask = W_dists[ir,0] .>= wk;
            C_sizes[ir,ic] = sumc( (W_dists[ir,0].*mask)' );
            //print /flush "wk = " wk "; Csize = " C_sizes[r,c];
        endfor;
    endfor;
    retp( C_sizes );
endp;

proc calc_MC(a,C_size);
    local dT,rv;

    dT = delta.^(T+1);
    rv = (1-dT).*(a.^(c-b)) - (delta-dT).*(C_size+w_h);
    retp( rv ./ (1-delta) );
endp;

proc calc_CR(a,C_size);
    local rv;

    rv = C_size+w_h-a.^(c-b);
    retp( rv.*(1-delta.^T) );
endp;

// called by the optimizer; assumes cur_Csize is set
// to size of coalition with smallest member cur_wk
// checks if (CR) and (MC) are satisfied by this value of a
cur_Csize = -1;
cur_wk = -1;
proc ineqp(x);
    local a,ws,wss;

    a = x[1];
    ws = calc_MC(a,cur_Csize);
    wss = calc_CR(a,cur_Csize);
    if( __DebugExtraStuff );
        format /rd 12,10; print /flush "\n---------\nws = " ws "\nwss = " wss "\nx = " a;
    endif;
    retp( cur_wk - max(ws,wss) );
endp;

proc hpay(x);
    local a;

    a = x[1];
    if( 0 == a ); retp (0); endif;
    // since co is minimizing, return -pay
    retp( -((cur_Csize+w_h)*a^b - a^c) );
endp;

// prints text s followed by n followed by ': ', no \n
// only if __DebugOutput is nonzero
proc(0)=message_header(s,n); if( __DebugOutput ); print /flush s n "\b: ";; endif; endp;
proc(0)=message(s); if( __DebugOutput ); print /flush s " ";; endif; endp;

/*
** returns 0 if no solution exists or a>=1 which admits
** a solution for a given coalition size and smallest wi
*/
proc solution_exists(minwi,Csize);
    local dt,wt,a,X;

    dT = delta^T;
    X = (1-dT)*(Csize+w_h)/(2-delta-dT);
    
    // MC and CR intersect
    if( X >= 1 ); 
        wt = (1-delta)*X;
        if( minwi >= wt );
            retp( X^(1/(c-b)) );
        else;
            retp( 0 );
        endif;
    endif; 

    // MC is only relevant
    wt = calc_MC(1,Csize); 
    retp( minwi >= wt );
endp;

// expects 1xN size distribution of N states
// returns (a,C,wk,pay) for each hegemon size in wh_sizes
proc(4)=demand_coalition(W_dist);
    local C_sizes,start,a,pay,g,ret,idx,
          opt_a,opt_pay,eq_a,eq_pay,eq_Csize,eq_wk;

    // set up optimizer
    _co_DirTol = 1e-6;
    _co_IneqProc = &ineqp;
    _co_Bounds = { 1 1e256 };
    __output = 0;
    trap 4;
    start = { 1 };

    C_sizes = coalitions(W_dist);

    // set globals for optimization
    opt_pay = zeros(cols(W_dist),1);
    opt_a = opt_pay;
    cur_wk = -1;
    // find a, if it exists, for all feasible coalitions
    for k (1, cols(W_dist), 1);
        if( cur_wk == W_dist[k] );
            continue;   // already done this min size
        endif;
        cur_wk = W_dist[k];
        cur_Csize = C_sizes[k];
        print /flush "wt =" cur_wk "\b; C_size =" cur_Csize;;

        // make sure optimizer starts at a place which admits a solution
        a = solution_exists(cur_wk,cur_Csize);
        if( a );
            start = a;
            {a,pay,g,ret} = co(&hpay,start);
            if( ret /= 0 ); 
                print /flush "Optimizer failed."; end;
            else;
                opt_a[k] = a;
                opt_pay[k] = -pay;
                print /flush "\n\t   a       C     pay_h  Q(a,C)    MC(a)   CR(a)\n\t"                  
                    opt_a[k] cur_Csize opt_pay[k] ((cur_Csize+w_h)*opt_a[k]^b)
                    calc_MC(opt_a[k],cur_Csize) calc_CR(opt_a[k],cur_Csize);
            endif;
        else;
            // no solution
            print /flush "--> no solution exists";
        endif;
    endfor;

    // get optimal a with highest payoff for hegemon
    idx = maxindc(opt_pay);
    eq_a = opt_a[idx];
    eq_pay = opt_pay[idx];
    eq_Csize = C_sizes[idx];
    eq_wk = W_dist[idx];
    //print /flush "wh = " w_h ": " eq_a[h] eq_pay[h] eq_wk[h] eq_Csize[h];

    retp(eq_a,eq_Csize,eq_wk,eq_pay);
endp;

proc(0)=calc_world(W_dist);
    local C_sizes,X_vals,min_C,mask,wts,yes,MC_min,CR_max,
          MC_atX,a_atX,cur_minw,a,Csize,wk,hegpay,sep;

    sep = "-------------------------------------------------------------------------------";

    print /flush "\n" sep "\nWorld: " W_dist "\nSize : " sumc(W_dist') "\n" sep;

    C_sizes = coalitions(W_dist);
    X_vals = (1-delta.^T).*(C_sizes+w_h)./(2-delta-delta.^T);

    min_C = (1-delta)/(1-delta^T)+1-w_h;
    mask = C_sizes .>= min_C;
    wts = ((1-delta).*X_vals).*mask + calc_MC(1,C_sizes).*(mask.<1);
    yes = W_dist .> wts;
    MC_min = calc_MC(1,C_sizes);
    CR_max = calc_CR(1,C_sizes);
    MC_atX = zeros(1,cols(W_dist));
    a_atX = MC_atX;

    for i (1, cols(W_dist), 1);
        if( mask[i] > 0 );
            // get MC at intersection
            a_atX[i] = X_vals[i]^(1/(c-b));
            MC_atx[i] = calc_MC(a_atX[i],C_sizes[i]);
        endif;
    endfor;

    print /flush "  min_w  C_size     X     wt(C)  intsc?  admit?   MC(1)  CR(1)   MC(atX) a(atX)";
    cur_minw = -1;
    for i (1, cols(W_dist), 1);
        if( cur_minw /= W_dist[i] );
            cur_minw = W_dist[i];
            print /flush W_dist[i]~C_sizes[i]~X_vals[i]/*
                  */~wts[i]~mask[i]~yes[i]~MC_min[i]~CR_max[i]/*
                  */~MC_atX[i]~a_atX[i];
        endif;
    endfor;
    print /flush //"\n-> min C_size to intersect: " min_C
                 "\n----< OPTIMAL (a,C) FOR ADMISSIBLE COALITIONS >----";

    {a,Csize,wk,hegpay}=demand_coalition(W_dist);

    print /flush "\n=> SPE: (a*,C*) = (" a  "\b," Csize "\b); "
             "min_w =" wk "\b; hpay = " hegpay "\n" sep;
    /*
    // show the extreme constraints for this world
    local which_r,mc,cr;
    a = seqas(1,3,50);
    which_r = cols(C_sizes);
    mc = calc_MC(a,C_sizes[1])~calc_MC(a,C_sizes[which_r]);
    cr = calc_CR(a,C_sizes[1])~calc_CR(a,C_sizes[which_r]);
    _plegstr = "MC(min)\000MC(max)\000CR(min)\000CR(max)";
    _plegctl = 1;
    _plwidth = 15;
    ytics(-0.5,3.5,0.5,0);
    xy(a,mc~cr);
    */
endp;

// v is 1xN vector; returns 1xK (K<=N) vector containing
// only the positive numbers in v
proc raw2world(v);
    retp( packr( miss(v,v.*(v.<=0))' )' );
endp;

/*
**********************************************************************
** MAIN
**********************************************************************
*/
format /rd 7,4;

// since we're dealing with a matrix, we can't have
// different number of elements across rows; so we
// pad each row with -1 (invalid state size); before
// processing use the raw2world() function to extract
// valid sizes, which then can be passed to calc_world()
worlds = 
{ 
  // broader/deeper trade-off everywhere
    0.02 0.07 0.25 0.85 0.85,
  // b/d trade-off shows up sporadically
    0.02 0.17 0.25 0.75 0.85,
  // exactly the reverse of b/d trade-off
    0.12 0.17 0.25 0.75 0.75
};
cls;
print /flush "\nPARAMETERS: "
    "\n\t     b =" b
    "\n\t     c =" c
    "\n\t     T =" T
    "\n\t delta =" delta
    "\n\t   w_h =" w_h;

for i (1, rows(worlds), 1 );
    calc_world( raw2world(worlds[i,0]) );
endfor;
