/*
** Copyright (c) 2006 Branislav L. Slantchev (slantchev@ucsd.edu)
*/
new;
library co;
coset;
eqSolveset;

#include params.dec;

__DebugOutput = 0;  // if /= 0, print optimization messages
__DebugExtraStuff = 0; // if /=0, print intermediate calls to optimizer

b = 0.90;
c = 1.15;
T = 5;
delta = 0.85;

N_DISTS_PER_SYSTEM = 2000; // draws per system type (more=better estimates)

N_BETA_VALUES = 25;  // # of beta values between 3 and 1
alpha = 1;
betas = seqase(3,1,N_BETA_VALUES);
string sims_fname = "sims";

// take care that max_wi does not exceed w_h
min_wi = 0.010;  // smallest potential contributor size
max_wi = 0.649;  // largest potential contributor size
min_wh = 0.650;  // hegemon size greater than largest other state
max_wh = 1.000;  // max value for hegemon size to use (used to compute X)
wh_sizes = seqase(min_wh,max_wh,N_HEGEMON_SIZES);

// the sum of contributors must equal WORLD_SIZE or else we're
// dealing with worlds of different wealth (which is OK but
// not the point of the simulations); we want to keep the
// wealth the same and simply distribute it differently;
// also, since we want to make sure distributions of many
// similarly-sized states are different from skewed ones, 
// make sure that for low values of beta, the number of states
// in the system is close to MAX_STATES_PER_DIST, and for
// higher values of beta, it is around MIN_STATES_PER_DIST
MIN_STATES_PER_DIST = 10;  // min number of states in each system
MAX_STATES_PER_DIST = 18;  // max number of states in each system
//WORLD_SIZE = 2*max_wh;

// beta starts high (skewed dist), so use few large states
// as beta decreases (uniform dist), nStates should increase
nStates = round(seqas(MAX_STATES_PER_DIST,MIN_STATES_PER_DIST,N_BETA_VALUES));

// rescale and sort all x values to be between minv and maxv
// can handle KxN matrix assuming it represents K distributions
// of N states each; we assume that x is distributed randomly
// by the beta distribution, and so takes on values between 0 and 1
// we use these values as anchors for the min/max; 
// finally, make sure the sum of sizes equals world_size
proc rescale(x,minv,maxv);
    local df,y,cx;

    // we want to find coeffs a,b for f(x)=ax+b such that
    // f(0) = minv and f(1) = maxv
    // that is, b=minv and a=(maxv-minv)
    df = maxv-minv;
    y = zeros(rows(x),cols(x));
    for r (1,rows(x),1);
        y[r,0] = df .* x[r,0] + minv;
        y[r,0] = sortc(y[r,0]',1)';
        cx = WORLD_SIZE / sumc(y[r,0]');
        y[r,0] = y[r,0] .* cx;
    endfor;
    retp(y);
endp;

// takes KxN matrix of W_dists and returns KxN matrix
// of coalition sizes such that if wk = W_dists[r,c], then
// C = retval[r,c], where C is sum of states with size >= wk
// given state system r 
proc coalitions(W_dists);
    local C_sizes,mask,wk;

    C_sizes = zeros(rows(W_dists),cols(W_dists));
    for ir (1, rows(W_dists), 1);    
        for ic (1, cols(W_dists), 1);
            // set wk and pick out wi >= wk only
            wk = W_dists[ir,ic];
            mask = W_dists[ir,0] .>= wk;
            C_sizes[ir,ic] = sumc( (W_dists[ir,0].*mask)' );
        endfor;
    endfor;
    retp( C_sizes );
endp;

proc contrib_constraint(a,C_size,w_h);
    local dT,rv;

    dT = delta.^(T+1);
    rv = (1-dT).*a.^(c-b) - (delta-dT).*(C_size+w_h);
    retp( rv ./ (1-delta) );
endp;

proc credib_constraint(a,C_size,w_h);
    local rv;

    rv = C_size+w_h-a.^(c-b);
    retp( rv.*(1-delta.^T) );
endp;

// called by the optimizer; assumes cur_Csize is set
// to size of coalition with smallest member cur_wk
// checks if (CR) and (MC) are satisfied by this value of a
cur_Csize = -1;
cur_wk = -1;
cur_wh = -1;
proc ineqp(x);
    local a,ws,wss;

    a = x[1];
    ws = contrib_constraint(a,cur_Csize,cur_wh);
    wss = credib_constraint(a,cur_Csize,cur_wh);
    if( __DebugExtraStuff );
        format /rd 12,10; print /flush "\n---------\nws = " ws "\nwss = " wss "\nx = " a;
    endif;
    retp( cur_wk - max(ws,wss) );
endp;

proc hpay(x);
    local a;

    a = x[1];
    if( 0 == a ); retp (0); endif;
    // since co is minimizing, return -pay
    retp( -((cur_Csize+cur_wh)*a^b - a^c) );
endp;

// prints text s followed by n followed by ': ', no \n
// only if __DebugOutput is nonzero
proc(0)=message_header(s,n); if( __DebugOutput ); print /flush s n ": ";; endif; endp;
proc(0)=message(s); if( __DebugOutput ); print /flush s " ";; endif; endp;

/*
** returns 0 if no solution exists or a>=1 which admits
** a solution for a given coalition size and smallest wi
*/
proc solution_exists(minwi,Csize,w_h);
    local dT,wt,a,X;

    dT = delta^T;
    X = (1-dT)*(Csize+w_h)/(2-delta-dT);
    
    // MC and CR intersect
    if( X >= 1 ); 
        wt = (1-delta)*X;
        if( minwi >= wt );
            retp( X^(1/(c-b)) );
        else;
            retp( 0 );
        endif;
    endif; 

    // MC is only relevant
    wt = contrib_constraint(1,Csize,w_h); 
    retp( minwi >= wt );
endp;

// expects 1xN size distribution of N states
// returns (a,C,wk,pay) for each hegemon size in wh_sizes
proc(4)=demand_coalition(W_dist);
    local C_sizes,start,a,pay,g,ret,n_wh,idx,nr,
          opt_a,opt_pay,eq_a,eq_pay,eq_Csize,eq_wk;

    // set up optimizer
    _co_DirTol = 1e-6;
    _co_IneqProc = &ineqp;
    _co_Bounds = { 1 1e256 };
    __output = 0;
    _co_Algorithm = 2;
    trap 4;

    C_sizes = coalitions(W_dist);
    n_wh = rows(wh_sizes);
    eq_a = zeros(n_wh,1);
    eq_pay = eq_a;
    eq_Csize = eq_a;
    eq_wk = eq_a;

    for h (1, n_wh, 1);
        // set globals for optimization
        cur_wh = wh_sizes[h];
        message("\n\n-----[ wh = "); message(cur_wh); message("]-------");
        opt_pay = zeros(cols(W_dist),1);
        opt_a = opt_pay;
        // find a, if it exists, for all feasible coalitions
        for k (1, cols(W_dist), 1);
            cur_wk = W_dist[k];
            cur_Csize = C_sizes[k];
            message_header("\nTrying wk = ", cur_wk);
            message_header("C_size = ", cur_Csize);

            // make sure optimizer starts at a place which admits a solution
            start = solution_exists(cur_wk,cur_Csize,cur_wh);
            if( start );
                //message(start);
                {a,pay,g,ret} = co(&hpay,start);
                if( ret /= 0 ); 
                    print /flush "Optimizer failed."; end;
                else;
                    opt_a[k] = a;
                    opt_pay[k] = -pay;
                    message(" --> (a,C,pay) =");
                    message(opt_a[k]); message(cur_Csize); message(opt_pay[k]);
                endif;
            else;
                // no solution
                message(" --> no solution exists");
            endif;
        endfor;
        // get optimal a with highest payoff for hegemon
        idx = maxindc(opt_pay);
        eq_a[h] = opt_a[idx];
        eq_pay[h] = opt_pay[idx];
        eq_Csize[h] = C_sizes[idx];
        eq_wk[h] = W_dist[idx];
        //print /flush "wh = " cur_wh ": " eq_a[h] eq_pay[h] eq_wk[h] eq_Csize[h];
    endfor;
    retp(eq_a,eq_Csize,eq_wk,eq_pay);
endp;

/*
** progress indicator with its own globals
*/
__progress_step_size = 10;  // show at 10% increments
__progress_job_size = -1;
__progress_last_step = -1;  // last step shown (in step_size incr)
__progress_display = -1;

proc(0)=set_progress(job_size);
    __progress_job_size = job_size;
    __progress_last_step = 0;
    __progress_display = 0;
    print /flush "0%..";;
endp;

proc(0)=update_progress(cur_step);
    local pct;
    pct = cur_step/__progress_job_size;
    if( pct >= 1 ); print /flush "100%";
    elseif( 100*(pct - __progress_last_step) >= __progress_step_size );
        __progress_display = __progress_display + 100*(pct - __progress_last_step);
        __progress_last_step = pct;
        print /flush ftos(__progress_display,"%0*.*lf",2,0) "%..";;
    endif;
endp;

/*
**********************************************************************
** MAIN
**********************************************************************
*/
print "Starting on " datestr(date()) " at " timestr(time());

// make sure random draws are replicable
par = sysstate(28,0);
par[1] = 16071973;
call sysstate(28,par);

format /rd 7,4;

eq_a = zeros(N_DISTS_PER_SYSTEM,N_HEGEMON_SIZES);
eq_Csize = eq_a;
eq_wk = eq_a;
eq_hpay = eq_a;

let vnames = beta wh a a_LB a_UB Csize Csize_LB Csize_UB hpay hpay_LB hpay_UB;
create fh = ^sims_fname with ^vnames, 0, 8;

print /flush "World wealth: " WORLD_SIZE;

for i (1,rows(betas),1); // for each system type
    print /flush ftos(i,"*.*lf",2,0) "/" ftos(rows(betas),"*.*lf",2,0) 
        " (beta = " ftos(betas[i],"*.*lf",4,2) 
        ", states = " ftos(nStates[i],"*.*lf",2,0) ": ";;

    // create N_DISTS_PER_SYSTEM distributions of nStates states
    W_dists = rndbeta(N_DISTS_PER_SYSTEM,nStates[i],alpha,betas[i]);
    W_dists = rescale(W_dists,min_wi,max_wi);
    
    set_progress(N_DISTS_PER_SYSTEM);   
    for r (1, N_DISTS_PER_SYSTEM, 1); // for each distribution of this system type
        update_progress(r);
        world = W_dists[r,0];
        if( sumc(world') - WORLD_SIZE > __macheps*10 );
            print /flush "ERROR: world wealth has changed at " r ": " sumc(world');
            end;
        endif;
        // calculate optimal demand/coalition; note that we get
        // one (a,C) for each hegemon size, so store all of them
        {a,Csize,wk,hegpay}=demand_coalition(world);
        eq_a[r,0] = a';
        eq_Csize[r,0] = Csize';
        eq_wk[r,0] = wk';
        eq_hpay[r,0] = hegpay';
    endfor;

    sys_a = zeros(N_HEGEMON_SIZES,3);    
    sys_Csize = zeros(N_HEGEMON_SIZES,3);    
    sys_hpay = zeros(N_HEGEMON_SIZES,3);    
  
    // we now have (a,C) for each simulated distribution for this 
    // system type; calculate the mean and confidence intervals;
    // this gives us a prediction for the system type and each 
    // hegemon size (eq_?? are N_DISTS_PER_SYSTEM x N_HEGEMON_SIZES)
    // results are N_HEGEMON_SIZES x 1 vectors with the means
    sys_a[0,1] = meanc(eq_a);
    sys_Csize[0,1] = meanc(eq_Csize);
    sys_hpay[0,1] = meanc(eq_hpay);

    percentile = 5; // which percentile do we want for CI
    which_lower = max(1,round(percentile*N_DISTS_PER_SYSTEM/100));
    which_upper = round((100-percentile)*N_DISTS_PER_SYSTEM/100);
    for j (1, cols(eq_a), 1 ); // take the confidence intervals
        // for each hegemon size, sort the values and take the
        // 'percentile' and '100-percentile' to get the CI bounds
        // results are N_HEGEMON_SIZES x 2 matrices (lower bound in
        // first column, upper bound in second column)
        z = sortc(eq_a[0,j],1);
        sys_a[j,2:3] = (z[which_lower]~z[which_upper]);
        z = sortc(eq_Csize[0,j],1);
        sys_Csize[j,2:3] = (z[which_lower]~z[which_upper]);
        z = sortc(eq_hpay[0,j],1);
        sys_hpay[j,2:3] = (z[which_lower]~z[which_upper]);
    endfor;

    // save results in N_HEGEMON_SIZES x 10 matrix with columns:
    // w_h, a (mean,LB,UB), C_size (mean,LB,UB), pay (mean,LB,UB)
    // append to the results file and save the value of beta too
    bv = betas[i].*ones(N_HEGEMON_SIZES,1);
    if( N_HEGEMON_SIZES /= writer(fh,bv~wh_sizes~sys_a~sys_Csize~sys_hpay) );
        errorlog "I/O error";
        fh = close(fh);
        end;
    endif;
endfor;
closeall fh;
print /flush "Done on " datestr(date()) " at " timestr(time());

